import { Markdown } from './Markdown';
import { OpenAPITabs, OpenAPITabsList, OpenAPITabsPanels } from './OpenAPITabs';
import { StaticSection } from './StaticSection';
import { generateSchemaExample } from './generateSchemaExample';
import { json2xml } from './json2xml';
import { stringifyOpenAPI } from './stringifyOpenAPI';
import { checkIsReference, createStateKey, resolveDescription } from './utils';
/**
 * Display an example of the response content.
 */
export function OpenAPIResponseExample(props) {
    var data = props.data, context = props.context;
    // if there are no responses defined for the operation
    if (!data.operation.responses) {
        return null;
    }
    var responses = Object.entries(data.operation.responses);
    // Sort response to get 200, and 2xx first
    responses.sort(function (_a, _b) {
        var a = _a[0];
        var b = _b[0];
        if (a === 'default') {
            return 1;
        }
        if (b === 'default') {
            return -1;
        }
        if (a === '200') {
            return -1;
        }
        if (b === '200') {
            return 1;
        }
        return Number(a) - Number(b);
    });
    var tabs = responses.map(function (_a) {
        var key = _a[0], responseObject = _a[1];
        var description = resolveDescription(responseObject);
        if (checkIsReference(responseObject)) {
            return {
                key: key,
                label: key,
                body: (<OpenAPIExample example={getExampleFromReference(responseObject)} context={context} syntax="json"/>),
                footer: description ? <Markdown source={description}/> : undefined,
            };
        }
        if (!responseObject.content || Object.keys(responseObject.content).length === 0) {
            return {
                key: key,
                label: key,
                body: <OpenAPIEmptyResponseExample />,
                footer: description ? <Markdown source={description}/> : undefined,
            };
        }
        return {
            key: key,
            label: key,
            body: <OpenAPIResponse context={context} content={responseObject.content}/>,
            footer: description ? <Markdown source={description}/> : undefined,
        };
    });
    if (tabs.length === 0) {
        return null;
    }
    return (<OpenAPITabs stateKey={createStateKey('response-example')} items={tabs}>
            <StaticSection header={<OpenAPITabsList />} className="openapi-response-example">
                <OpenAPITabsPanels />
            </StaticSection>
        </OpenAPITabs>);
}
function OpenAPIResponse(props) {
    var context = props.context, content = props.content;
    var entries = Object.entries(content);
    var firstEntry = entries[0];
    if (!firstEntry) {
        throw new Error('One media type is required');
    }
    if (entries.length === 1) {
        var mediaType = firstEntry[0], mediaTypeObject = firstEntry[1];
        return (<OpenAPIResponseMediaType context={context} mediaType={mediaType} mediaTypeObject={mediaTypeObject}/>);
    }
    var tabs = entries.map(function (entry) {
        var mediaType = entry[0], mediaTypeObject = entry[1];
        return {
            key: mediaType,
            label: mediaType,
            body: (<OpenAPIResponseMediaType context={context} mediaType={mediaType} mediaTypeObject={mediaTypeObject}/>),
        };
    });
    return (<OpenAPITabs stateKey={createStateKey('response-media-types')} items={tabs}>
            <StaticSection header={<OpenAPITabsList />} className="openapi-response-media-types">
                <OpenAPITabsPanels />
            </StaticSection>
        </OpenAPITabs>);
}
function OpenAPIResponseMediaType(props) {
    var mediaTypeObject = props.mediaTypeObject, mediaType = props.mediaType;
    var examples = getExamplesFromMediaTypeObject({ mediaTypeObject: mediaTypeObject, mediaType: mediaType });
    var syntax = getSyntaxFromMediaType(mediaType);
    var firstExample = examples[0];
    if (!firstExample) {
        return <OpenAPIEmptyResponseExample />;
    }
    if (examples.length === 1) {
        return (<OpenAPIExample example={firstExample.example} context={props.context} syntax={syntax}/>);
    }
    var tabs = examples.map(function (example) {
        return {
            key: example.key,
            label: example.example.summary || example.key,
            body: (<OpenAPIExample example={example.example} context={props.context} syntax={syntax}/>),
        };
    });
    return (<OpenAPITabs stateKey={createStateKey('response-media-type-examples')} items={tabs}>
            <StaticSection header={<OpenAPITabsList />} className="openapi-response-media-type-examples">
                <OpenAPITabsPanels />
            </StaticSection>
        </OpenAPITabs>);
}
/**
 * Display an example.
 */
function OpenAPIExample(props) {
    var example = props.example, context = props.context, syntax = props.syntax;
    var code = stringifyExample({ example: example, xml: syntax === 'xml' });
    if (code === null) {
        return <OpenAPIEmptyResponseExample />;
    }
    return context.renderCodeBlock({ code: code, syntax: syntax });
}
function stringifyExample(args) {
    var example = args.example, xml = args.xml;
    if (!example.value) {
        return null;
    }
    if (typeof example.value === 'string') {
        return example.value;
    }
    if (xml) {
        return json2xml(example.value);
    }
    return stringifyOpenAPI(example.value, null, 2);
}
/**
 * Get the syntax from a media type.
 */
function getSyntaxFromMediaType(mediaType) {
    if (mediaType.includes('json')) {
        return 'json';
    }
    if (mediaType === 'application/xml') {
        return 'xml';
    }
    return 'text';
}
/**
 * Get examples from a media type object.
 */
function getExamplesFromMediaTypeObject(args) {
    var _a;
    var _b, _c;
    var mediaTypeObject = args.mediaTypeObject, mediaType = args.mediaType;
    if (mediaTypeObject.examples) {
        return Object.entries(mediaTypeObject.examples).map(function (_a) {
            var key = _a[0], example = _a[1];
            return {
                key: key,
                example: checkIsReference(example) ? getExampleFromReference(example) : example,
            };
        });
    }
    if (mediaTypeObject.example) {
        return [{ key: 'default', example: { value: mediaTypeObject.example } }];
    }
    if (mediaTypeObject.schema) {
        if (mediaType === 'application/xml') {
            // @TODO normally we should use the name of the schema but we don't have it
            // fix it when we got the reference name
            var root = (_c = (_b = mediaTypeObject.schema.xml) === null || _b === void 0 ? void 0 : _b.name) !== null && _c !== void 0 ? _c : 'object';
            return [
                {
                    key: 'default',
                    example: {
                        value: (_a = {},
                            _a[root] = generateSchemaExample(mediaTypeObject.schema, {
                                xml: mediaType === 'application/xml',
                            }),
                            _a),
                    },
                },
            ];
        }
        return [
            {
                key: 'default',
                example: { value: generateSchemaExample(mediaTypeObject.schema) },
            },
        ];
    }
    return [];
}
/**
 * Empty response example.
 */
function OpenAPIEmptyResponseExample() {
    return (<pre className="openapi-response-example-empty">
            <p>No body</p>
        </pre>);
}
/**
 * Generate an example from a reference object.
 */
function getExampleFromReference(ref) {
    return { summary: 'Unresolved reference', value: { $ref: ref.$ref } };
}
